#ifndef _layout0_h
#define _layout0_h

#include "node.h"

typedef struct level {
  node *nodes;
  int numberOfNodes;
  float maxPos, minPos;
  node placedNodes;
} level;

extern int numberOfLevels;
extern level* levels;

extern list selectPreds(node);
extern list selectSuccs(node);
extern node selectFrom(edge);
extern node selectTo(edge);

extern int cmpDreamPositions(const void*, const void*);
extern void downwardDreamPos();
extern void upwardDreamPos();
extern void insertLevelIndex(int);
extern void verboseShowLevel(int);

extern int isDummy(node);
extern node newDummyNode(int);

extern int priorityOfNode(node);
extern float nodeDistance(node, node);

extern float dummy2DummyDistance, dummy2NodeDistance,
             node2NodeDistance, node2DummyDistance;

#define PHASES 6

extern int phaseIterations[];
extern int phaseFirstUp[];
extern int phaseOnlyUp[];
extern int phaseOnlyDown[];
extern void(*phaseBodyUp[])(int);
extern void(*phaseBodyDown[])(int);

extern void initPhases();

extern int global_iterations;
extern void generateStatistic();
extern void reportStatistics(int,int);

extern int maxPriority, dummyFactor;
extern int useStatistics;
extern int compactifyLevels;

/*---------------------------------------------------------------------------.
 | We have different methods of calculating the dreamposition of a node. For |
 | example we can calculate the barycenter or the median of some nodes       |
 | related to the node for which we want to calculate the dream position.    |
 | First of all we can use its predecessors *and* succesor nodes. The second |
 | possibility is to use just one type of nodes such as predecessors or      |
 | successor. In this second case the arguments to the `calculate' function  |
 | choose the appropriate mode.                                              |
 `---------------------------------------------------------------------------*/
typedef struct DreamPosition {
  float (*get)(node);
  void (*calculate)(node);
  void (*calc1)(node);
  void (*calc2)(node);
  list (*get_edge_list)(node);
  node (*get_node)(edge);
} DreamPosition;

extern DreamPosition dreamPosition;

extern void useMedian();
extern void useBarycenter();
extern void useDoublePosition();
extern void useSinglePosition();

extern unsigned int edge_crossings(node,node);
extern void levelSweep(int);

extern void initNodeDistances();

extern int numberOfEdgeCrossings();

#endif
