#include "io.h"
#include "layout0.h"
#include "layout1.h"
#include "mem.h"
#include "macros.h"


#if 0

/* wrong try .... should think first */

/*---------------------------------------------------------------------------.
 | Here we do a modified breathfirst search to assign a level to each node.  |
 | The user may request a level and if the predecessors are assigned to a    |
 | level lower than the requested level then it becomes the actual level of  |
 | the node otherwise the maximum over the levels of the predecessors + 1    |
 | becomes the actual level. For this decision we have to know the levels of |
 | the predecessors. So the breath first search queue is searched until a    |
 | node is found which has all it's predecessors assigned a level.           |
 |                                                                           |
 | If we implement this in such a way that all nodes in the queue before     |
 | such a node are put back at the end of the queue this will mean quadratic |
 | time in the worst case.                                                   |
 |                                                                           |
 | So we use another algorithm. First of all we assign to the mark field of  |
 | each node the number of it's predecessors. When a node has been assigned  |
 | to a level than all the mark fields of its successors are decremented and |
 | only when this count reaches zero then this successor is put on the       |
 | breath first queue. So we have the invariant that as long as their are    |
 | non visited nodes (nodes that have not been assigned to a level) so their |
 | are always nodes on the queue. This invariant holds at the beginning      |
 | because the queue is initialized with the list of roots ...               |
 |                                                                           |
 | REM1: All lists of successors (predecessors) should be read as all lists  |
 | of successors (predecessor) edges.                                        |
 `---------------------------------------------------------------------------*/

static void adjusted_dream_level(node n)
{
  if(n->aux.dreamLevel!=-1)
    fprintf(stderr, WARNINGSTR
      "dreamLevel %d of node %s has to be set to %d\n",
      n->aux.dreamLevel, n->name, n->level);
}

static void set_to_max(node n, int max1, int max2)
{
  int pos = ((max1>max2) ? max1 : max2) + 1;
  int dl = n -> aux.dreamLevel;

  if(dl>=pos)
    n -> level = dl;
  else
    {
      n -> level = pos;
      adjusted_dream_level(n);
    }
}

int alloc_level_breath_first = 0;
void setLevelWithQueue_complicated()
{
  node n, n2;
  list l;
  rlist s = 0;

  for(n=nodes; n; n=cdr(n))
    n->mark = list_length(n->preds);

  for(l=roots; l; l=cdr(l))
    s = rlist_consAtEnd( car(l), s );

  while(s)
    {
      int max_succs, min_succs, max_preds, min_preds;
      int dl;

      n = (node)car(s);
      dl = n -> aux.dreamLevel;
      s = rlist_cdrAtFront(s);

      min_preds = max_preds = -1;
      for(l=n->preds; l; l=cdr(l))
        {
	  n2 = ((edge)car(l))->from;
	  if(n2 -> level != -1)
	    {
	      if(min_preds == -1 || n2 -> level < min_preds)
	        min_preds = n2 -> level;

	      if(n2 -> level > max_preds)
	        max_preds = n2 -> level;
	    }
	}

      min_succs = max_succs = -1;
      for(l=n->succs; l; l=cdr(l))
        {
	  n2 = ((edge)car(l))->to;
	  if(n2 -> level != -1)
	    {
	      if(min_succs == -1 || n2 -> level < min_succs)
	        min_succs = n2 -> level;

	      if(n2 -> level > max_succs)
	        max_succs = n2 -> level;
	    }
	}
      
      if(max_succs == -1 && max_preds == -1) /* defensive */
        {
	  n -> level = dl == -1 ? 0 : dl;
	}
      else
      if(max_succs == -1) /* this should be the most common case */
        {
	  if(dl == -1)
	    n -> level = max_preds + 1;
	  else
	  if(dl > max_preds)
	    n -> level = dl;
	  else
	    {
	      n -> level = max_preds + 1;
	      adjusted_dream_level(n);
	    }
	}
      else
      if(max_preds == -1) /* defensive */
        {
	  if(dl==-1)
	    {
	      if(min_succs>0)
	        n -> level = min_succs - 1;
	      else
	        n -> level = max_succs + 1;
	    }
	  else
	    {
	      if(dl > max_succs)
	        n -> level = dl;
	      else
	        {
		  n -> level = max_succs + 1;
		  adjusted_dream_level(n);
		}
	    }
	}

      else  /* now we have successors and predecessors */
      
      if(max_succs + 1 < min_preds )
        {
	  if(dl == -1)
	    n -> level = max_succs + 1;
	  else
	  if(max_succs < dl && dl < min_preds)
	    n -> level = dl;
	  else set_to_max(n, max_preds, max_succs);
        }
      else
      if(max_preds + 1 < min_succs)
        {
	  if(dl==-1)
	    n -> level = max_preds + 1;
	  else
	  if(max_preds < dl && dl < min_succs)
	    n -> level = dl;
	  else set_to_max(n, max_preds, max_succs);
	}
      else set_to_max(n, max_preds, max_succs);

      for(l=n->succs; l; l=cdr(l))
        {
	  n2 = ((edge)car(l))->to;
	  if(!--n2->mark)                     /* decrement predecessor count */
	    {
	      if(alloc_level_breath_first)
	        s = rlist_consAtEnd(n2,s);
	      else
	        s = rlist_consAtFront(n2,s);
	    }
	}
    }
}
#endif

static int level_multiplicator = 1;

static void setLevelWithQueue()
{
  list l; rlist s = 0;

  for(l=roots; l; l=cdr(l))
    s = rlist_consAtEnd( car(l), s );

  while(s)
    {
      node n = (node) car(s);

      int max = -1;	/* maximum predecessor and succesors level */

      int hit_dream_Level = 0;

      s = rlist_cdrAtFront(s);

      if(n -> mark) continue;
      else n -> mark = 1;

      for(l=n->preds; l; l=l->next)
        {
	  node from = ((edge)car(l)) -> from;
	  if(from -> level != -1)
	    {
	      if(from -> level == n -> aux.dreamLevel) hit_dream_Level = 1;
	      if(from -> level > max) max = from -> level;
	    }
	}
      
      for(l=n->succs; l; l=l->next)
        {
	  node to = ((edge)car(l)) -> to;
	  if(to -> level != -1)
	    {
	      if(to -> level == n -> aux.dreamLevel) hit_dream_Level = 1;
	      if(to -> level > max) max = to -> level;
	    }
	}
      
      if(n -> aux.dreamLevel == -1)
        n -> level = max + 1;		/* at least 0 */
      else
      if(!hit_dream_Level)
        n -> level = n -> aux.dreamLevel;
      else
        {
	  n -> level = max + 1;
          fprintf(stderr, WARNINGSTR
            "dreamLevel %d of node %s has to be set to %d\n",
            n->aux.dreamLevel/level_multiplicator, n->name,
	    n->level/level_multiplicator);
	}

      for(l=n->preds; l; l=l->next)
        {
	  node from = ((edge)car(l)) -> from;
	  if(from -> level == -1)
	    s = rlist_consAtEnd(from, s);
	}
      
      for(l=n->succs; l; l=l->next)
        {
	  node to = ((edge)car(l)) -> to;
	  if(to -> level == -1)
	    s = rlist_consAtEnd(to, s);
	}
      
    }
}

static void flipEdgesAccordingToLevel()
{
  edge e;
  for(e = edges; e; e=cdr(e))
    if(e->from -> level > e->to->level)
      flipEdge(e);
}

static void copyLevelToDreamLevel()
{
  node n;
  for(n=nodes; n; n=cdr(n))
    {
      n -> aux.dreamLevel = n -> level;
      n -> level = -1;
      n -> mark = 0;
    }
}

/*--------------------------------------------------------------------*\
 | this is an buffer which is big enough to hold all nodes of a Level |
\*--------------------------------------------------------------------*/
lBufData *lBuf = 0;
lBufData *lBuf2 = 0;

static void generateLevelBuffer()
{
  int i,m;

  if(!numberOfLevels) return;

  for(i=m=0; i<numberOfLevels; i++)
    if(levels[i].numberOfNodes>m)
      m=levels[i].numberOfNodes;
  lBuf = (lBufData*) save_calloc(m, sizeof(lBufData));
  lBuf2 = (lBufData*) save_calloc(m, sizeof(lBufData));
}

void copy_level_to_buffer(int level)
{
  int i, m=levels[level].numberOfNodes;
  node* a=levels[level].nodes;

  for(i=0; i<m; i++)
    lBuf[i].asNode = a[i];
}

void copy_buffer_to_level(int level)
{
  int i, m=levels[level].numberOfNodes;
  node* a=levels[level].nodes;

  for(i=0; i<m; i++)
    a[i] = lBuf[i].asNode;
}

#ifndef USE_OLD_DOUBLE_EDGE_DETECTION
static int haveDoubleEdges()
{
  node n;

  for(n=nodes; n; n=cdr(n))
    {
      list k;

      for(k=n->succs; k; k=cdr(k))
        {
	   edge e1 = (edge)car(k);
	   node t = e1 -> to;
	   list l;
           
	   for(l=t->succs; l; l=cdr(l))
	     {
	       edge e2 = (edge) car(l);
	       node t2 = e2 -> to;
	       if(t2==n && e1!=e2) return 1;	/* defensive */
	     }

	   for(l=t->preds; l; l=cdr(l))
	     {
	       edge e2 = (edge) car(l);
	       node f2 = e2 -> from;
	       if(f2==n && e1!=e2) return 1;
	     }
	}
    }

  return 0;
}

static void splitAndCloneEdge(edge e)
{
  node n = newDummyNode(-1);
  edge e1, e2;

  e1 = newEdge(e -> from, n);
  e2 = newEdge(n, e -> to);
  
  e2 -> attr = e -> attr;

  correctDummyEdgesAttrs(e1, e2, 0);
}


static void generateDummyNodesForDoubleEdges()
{
  node n;
  edge e;

  for(e=edges; e; e=cdr(e))
    e->mark = 0;

  /*
   * mark all edges that should be split
   */

  for(n=nodes; n; n=cdr(n))
    {
      list k;

      for(k=n->succs; k; k=cdr(k))
        {
	   edge e1 = (edge)car(k);
	   node t = e1 -> to;
	   list l;
           
	   for(l=t->succs; l; l=cdr(l))
	     {
	       edge e2 = (edge) car(l);
	       node t2 = e2 -> to;
	       if(t2==n && e1!=e2)
	         e2->mark = 1;
	     }

	   for(l=t->preds; l; l=cdr(l))
	     {
	       edge e2 = (edge) car(l);
	       node f2 = e2 -> from;
	       if(f2==n && e1!=e2)
	         e2->mark = 1;
	     }
	}
    }

  for(e=edges; e; e=cdr(e))
    if(e->mark) splitAndCloneEdge(e);

  deleteMarkedEdges();
}

static void generateDummyLevels()
{
  node n;
  verbose("generating dummy levels ...");
  level_multiplicator *= 2;
  for(n=nodes; n; n=cdr(n))
    if(n->level!=-1) n -> level *= 2;
}
#endif

void setLevel()
{
#ifndef USE_OLD_DOUBLE_EDGE_DETECTION
  verbose("searching for double edges ...");
  if(haveDoubleEdges())
    {
      verbose("found double edges.");
      generateDummyLevels();
      generateDummyNodesForDoubleEdges();
    }
  else verbose("no double edges found.");
#endif

  verbose("mapping nodes to levels ...");
  copyLevelToDreamLevel();
  setLevelWithQueue();
  flipEdgesAccordingToLevel();
}

int numberOfLevels = 0;

#ifdef USE_OLD_DOUBLE_EDGE_DETECTION
int numberOfDummyLevels = 0;
#endif
level* levels = 0;

int normalizeLevelsFlag = 1;
void normalizeLevels()
{
  int i, nol;
  int *tol;	/* table of levels */
  int fzl;	/* free zero level */

  node n;

  int minLevel = MAXINT, maxLevel = -1;

  if(!nodes) return;

  /*------------------------------------------*\
   | search for mimum and maximum input level |
  \*------------------------------------------*/
  for(n=nodes; n; n=cdr(n))
    {
      int nl = n->level;

      if(nl>maxLevel)
        maxLevel=nl;

      if(nl<minLevel)
        minLevel=nl;
    }

  if(verboseFlag>1)
    fprintf(stderr, VERYVERBOSESTR "input levels are between %d and %d\n",
            minLevel, maxLevel);
  
  if(!normalizeLevelsFlag)
    {
      minLevel = 0;
      numberOfLevels = maxLevel + 1;
      return;
    }

  nol = maxLevel - minLevel + 1;

  tol = (int*) save_calloc(nol, sizeof(int));
  for(i=0; i<nol; i++)
    tol[i] = 0;

  for(n=nodes; n; n=cdr(n))	/* mark all used levels */
    tol[n->level-minLevel] = 1;

  for(i=0, fzl=0; i<nol; i++)	/* calculate absolute position of marked */
    if(tol[i]) tol[i] = fzl++;	/* levels */

  for(n=nodes; n; n=cdr(n))	/* normalize levels */
    n->level = tol[n->level - minLevel];

#ifdef USE_OLD_DOUBLE_EDGE_DETECTION
  for(i=0; i<nol; i++)
    tol[i] = 0;

  for(n=nodes; n; n=cdr(n))	/* mark all levels which need a dummy level */
    {
      list l;

      if(tol[n->level])
        {
          for(l=n->succs; l; l=cdr(l))
            {
	      edge e1 = (edge) car(l);
	      list k;
	      for(k=cdr(l); k; k=cdr(k))
	        {
	          edge e2 = (edge) car(k);
	          if( e1->to == e2->to &&
		      e1->to->level == e2->to->level &&
		      e1->to->level == n->level + 1 )
	            {
		      tol[n->level]=1;
		      goto testNextNode;
		    }
                }
	    }
        }
testNextNode:
    }

  for(i=0, fzl=0; i<nol; i++)
    if(tol[i])
      {
        tol[i]=fzl;
	numberOfDummyLevels++;
        fzl += 2;
      }
    else
      tol[i]=fzl++;

  for(n=nodes; n; n=cdr(n)) 
    n->level = tol[n->level];

  /*-----------------------------------------------------------*\
   | Skip empty levels and store real level position for each  |
   | level in tol. Generate also dummy levels for double edges |
   | between two nodes in adjacent levels.                     |
  \*-----------------------------------------------------------*/

  numberOfLevels = fzl;

  if(verboseFlag>1)
    {
      int dl = nol-fzl;
      int rl = fzl - numberOfDummyLevels;
      if(dl)
        fprintf(stderr, VERYVERBOSESTR "%d empty level%s deleted\n",
	        dl, dl == 1 ? "" : "s");
      fprintf(stderr, VERYVERBOSESTR "%d real level%s allocated\n",
              rl, rl == 1 ? "" : "s");
      fprintf(stderr, VERYVERBOSESTR "%d dummy level%s generated\n",
        numberOfDummyLevels, numberOfDummyLevels == 1 ? "" : "s");
    }
#else
  numberOfLevels = fzl;

  if(verboseFlag>1)
    fprintf(stderr, VERYVERBOSESTR "%d empty level%s deleted\n",
            nol-fzl, nol-fzl== 1 ? "" : "s");
#endif

  free(tol);
}

void allocateLevels()
{
  int i;
  node l;

  if(numberOfLevels)
    levels = (level*) save_calloc(numberOfLevels, sizeof(level));

  /*------------------------------------*\
   | Initialize numberOfNodes per level |
  \*------------------------------------*/

  for(i=0; i<numberOfLevels; i++)
    levels[i].numberOfNodes = 0;

  /*--------------------------*\
   | count nodes in one level |
  \*--------------------------*/

  for(l=nodes; l; l=cdr(l))
    levels[l->level].numberOfNodes++;

  /*-----------------------------------------*\
   | allocate array for nodes for all levels |
  \*-----------------------------------------*/

  for(i=0; i<numberOfLevels; i++)
    levels[i].nodes =
      (node*) save_calloc(levels[i].numberOfNodes, sizeof(node));

  /*--------------------------------------*\
   | ... and insert nodes in these arrays |
  \*--------------------------------------*/

  for(i=0; i<numberOfLevels; i++)
    levels[i].numberOfNodes = 0;

  for(l=nodes; l; l=cdr(l))
    (levels[l->level].nodes)[ levels[l->level].numberOfNodes++ ] = l;

  /*------------------------------------------------------------*\
   | here we should have the same number of nodes per level ;-) |
  \*------------------------------------------------------------*/

  generateLevelBuffer();
}

/*---------------------------------------------------------------------------.
 | For each edge and all levels between source and destination of this edge  |
 | dummy nodes are generated.                                                |
 `---------------------------------------------------------------------------*/
void generateDummies()
{
  edge l;
  int len = 0;

  verbose("generating dummies ...");

  for(l=edges; l; l=cdr(l))
    {
      node f, t;
      int reverse_l;
      
      if(l -> to -> level > l -> from -> level)
        {
          f = l->from;
	  t = l->to;
	  reverse_l = 0;
	}
      else
        {
	  f = l->to;
          t = l->from;
	  reverse_l = 1;
	}

      if( t->level - f->level > 1)
        {
          int i;
	  node prev, n=0;
	  edge e, first=0;
        
	  for(prev=f, i=f->level+1; i<t->level; i++, len++, prev=n)
	    {
	      n = newDummyNode(i);
	      e = newEdge(prev, n); /* D A N G E R O U S */
	      if(i==f->level+1)
	        first = e;
	      else
	        e->attr = fatherAttr(l->attr);
	    }
	  correctDummyEdgesAttrs(first, l, reverse_l);
          redirectSourceOfEdge(l, n, reverse_l);
        }
    }

  if(verboseFlag)
    {
      fprintf(stderr, VERBOSESTR
              "generated %d dummy node%s\n",
	      len, len == 1 ? "" : "s" );
    }
}

void setYposToLevelIndex()
{
  node n;
  for(n=nodes; n; n=cdr(n))
    n->ypos = n->level;
}
