#include "lists.h"

#include "mem.h"

CellCodeTemplate(list)

void list_freeList(list l)
{
  list n;

  for(; l; l=n)
    {
      n = cdr(l);
      list_freeCell(l);
    }
}

list list_cons(void*d,list l)
{
  list new = list_newCell();
  car(new) = d; cdr(new) = l;
  return new;
}

/*---------------------------------------------------------------------------.
 | this function deletes the head of the list an frees the list node         |
 `---------------------------------------------------------------------------*/
list list_removeHead(list l)
{
  list tmp = cdr(l);
  list_freeCell(l);
  return tmp;
}

/*---------------------------------------------------------------------------.
 | this function destroys it's first argument                                |
 `---------------------------------------------------------------------------*/
list list_append(list a, list b)
{
  list tmp;

  if(a)
    {
      for(tmp=a; cdr(tmp); tmp=cdr(tmp))
        ;
      cdr(tmp) = b;
      return a;
    }
  else
    {
      return b;
    };
}

LengthTemplate(list)

void list_remove(list *l, void* d)
{
  list *last = 0, p;

  if(*l)
    {
      for(last = l, p = *l; p ; last = &(*last)->next)
        if( car(p) == d )
	  {
	    list q = p;
	    *last = cdr(p); /* skip *p */
	    p = cdr(p);
	    list_freeCell(q);
	  }
	else
	  {
	    p = cdr(p);
	  }
    }
}

void list_reverse(list* lp)
{
  list l=0, n=*lp, tmp;
  while(n)
    {
      tmp = cdr(n); cdr(n)=l;
      l=n; n=tmp;
    }
  *lp=l;
}

list list_insertSorted(void*d,list lIn,int(*f)(void*,void*))
{
  list last, new, l=lIn;
  for(last=0; l && (*f)(d,car(l))<0; last = l, l=cdr(l))
    ;
  new = list_newCell();
  new->data = d; new->next = l;
  if(last)
    {
      last->next = new;
      return lIn;
    }
  else
    {
      return new;
    }
}

/*---------------------------------------------------------------------------.
 | we have the invariant that all rlist elements are refrenced (in the same  |
 | rlist) exactly twice                                                      |
 `---------------------------------------------------------------------------*/

CellCodeTemplate(rlist)

rlist rlist_consAtEnd(void* d, rlist l)
{
  rlist new = rlist_newCell();
  new->data = d;
  if(l)
    {
      new->next = l;
      new->prev = l->prev;
      l->prev->next = new;
      l->prev = new;
      return l;
    }
  else
    {
      new->next = new->prev = new;
      return new;
    }
}

rlist rlist_consAtFront(void * d, rlist l)
{
   return rlist_consAtEnd(d,l) -> prev;
}

CellCodeTemplate(dlist)

dlist dlist_insertSorted(void*d, dlist lIn, int(*f)(void*,void*))
{
  dlist l, last, new;

  for( last=0, l=lIn; l &&  (*f)(d, car(l)) < 0; last = l, l = cdr(l) )
    ;
  new = dlist_newCell();
  new->data = d; new->next = l; new->prev = last;
  if(l) l->prev = new;
  if(last)
    {
      last->next = new;
      return lIn;
    }
  else
    return new;
}

/*---------------------------------------------------------------------------.
 | this is destruktive                                                       |
 `---------------------------------------------------------------------------*/
rlist rlist_cdrAtFront(rlist l)       
{
  rlist res = 0;
  if(l!=l->next)
    {
      l->next->prev = l->prev;
      res = l->prev->next = l->next;
    }
  rlist_freeCell(l);
  return res;
}

void* rlist_carAtFront(rlist l) { return l->data; }

rlist rlist_insertSorted(void*d, rlist qIn, int(*f)(void*,void*))
{
  rlist q, start, new;

  if(qIn)
    {
      start = q = qIn;
      do
        {
	  if( (*f)(d, car(q)) >= 0 )
	    break;
	  q = cdr(q);
	}
      while(q != start);
      new = rlist_newCell();
      new->data = d;
      new->next = q; new->prev = q->prev;
      q->prev = new; new->prev->next = new;
      return qIn;
    }
  else
    {
      new = rlist_newCell();
      new->data = d;
      new->next = new->prev = new;
      return new;
    }
}
