#ifndef _lists_h
#define _lists_h

/*---------------------------------------------------------------------------.
 | simple lists part                                                         |
 `---------------------------------------------------------------------------*/

typedef struct _list {
  void* data;
  struct _list* next;
} _list, *list;

#define car(l) (l->data)
#define cdr(l) (l->next)

extern void list_freeCell(list);
extern void list_freeList(list);
extern list list_removeHead(list);

extern list list_cons(void*,list);
extern list list_append(list,list);
extern void list_remove(list*, void*);
extern void list_reverse(list*);
extern list list_insertSorted(void*,list,int(*)(void*,void*));

extern void list_forallDo(list,void(*f)(void*));
extern int  list_length(list);
extern void list_freeList(list);

#define LengthTemplate(t)\
int t ## _length( t l )\
{\
  int res = 0;\
  for(; l; l=cdr(l), res++)\
    ;\
  return res;\
}\

/*---------------------------------------------------------------------------.
 | Double linked lists with zero predecessor at head of list and zero        |
 | successor at end of. This enables us two use the same iterator code for a |
 | dlist as with a simple list (list).                                       |
 `---------------------------------------------------------------------------*/

typedef struct _dlist {
  void* data;
  struct _dlist* prev;
  struct _dlist* next;
} _dlist, *dlist;

extern dlist dlist_insertSorted(void*, dlist, int(*)(void*,void*));

/*---------------------------------------------------------------------------.
 | Double linked lists with ring structure. The advantage is that append     |
 | operations use constant time. The disadvantage is that we can not use the |
 | same iterator code as with dlist's.                                       |
 `---------------------------------------------------------------------------*/

typedef dlist rlist;
typedef _dlist _rlist;

extern rlist rlist_consAtEnd(void*,rlist);
extern rlist rlist_consAtFront(void*,rlist);
extern rlist rlist_cdrAtFront(rlist);
extern void* rlist_carAtFront(rlist);
extern rlist rlist_insertSorted(void*,rlist,int(*)(void*,void*));

#endif
