#ifndef _node_h
#define _node_h

#include "attr.h"
#include "lists.h"
#include "labels.h"

typedef struct _node {

  char* name;			/* for dummy nodes and when checking */
  label label;

  struct _node *next;		/* for list of all nodes */

  list succs, preds;		/* List of edges with attributes */

  float xpos, ypos;	

  float distance;

#ifdef STATISTICS
  float xxpos, xxxpos;
#endif

  struct _attribute *attr;

  int level;			/* real level (-1 means undefined) */

  int mark;			/* mark field for traversal && level index */

  struct _node * scc;	/* representant of strongly connected component */

  union {
    int onThePath;		/* used for the detection of backedges */
    int minimal_back_dpi;	/* used for SCC generation */
    int priority;		/* also used in phase 3 */
    unsigned int id;		/* used for printing */
    int dreamLevel;		/* input level */

    struct _node *nextPlaced;	/* neighbour in phase 3 */

    struct _node * median;	/* stores dreamPositions */

    float bc;
  } aux;

} _node, *node;

extern node nodes;
extern list roots;

extern int node_length(node);

extern void printNodeWithSuccessors(node);
extern void printAllNodes();
extern void resetMarkFields();
extern void reverseAllNodes();
extern int checkGraph();


typedef struct _edge {
  node from, to;
  float fx, fy, tx, ty;
  int mark;
  struct _edge *next;
  struct _attribute *attr;
} _edge, *edge;

/* forward definition conflict */
extern void correctDummyEdgesAttrs(edge,edge,int);

extern edge edges;

extern node newNode(char*);
extern void disposeNode(node);
extern edge newEdge(node, node);
extern void deleteEdge(edge);  /* be very cautios when using this function */
extern void deleteMarkedEdges();
extern void redirectSourceOfEdge(edge,node,int);
extern void flipEdge(edge);


#define FROM_EDGE 0
#define TO_EDGE 1

#endif
