%{
#include "attr.h"
#include "labels.h"
#include "node.h"
#include "symbols.h"

extern int yyerror(const char *);
extern int yylex();
extern int yylineno;
%}

%union{
  char* str;
  symbol symb;
  node node;
  list list;
  label label;
  edge_parameters edgePara;
  int num;
  attribute attr;
};

%token LC
%token RC
%token KOMMA
%token COLON
%token LESSTHAN
%token GREATERTHAN
%token STRING

%token IDENT
%token LEVEL

%type <symb> IDENT
%type <node> nodeAndAttr
%type <list> edgeDefList attrComponentsList
%type <label> labelDef
%type <edgePara> edgeDef
%type <str> STRING
%type <num> direction levelDef LEVEL
%type <attr> attributeRef

%%
start
:
  defList
;

defList
:
  defList nodeDef
|
  defList attrDef
|
  /* epsilon */
;

nodeDef
:
  nodeAndAttr edgeDefList
	{ partialInsertList($1,$2); }
;

nodeAndAttr
:
  IDENT levelDef attributeRef labelDef
	{ 
	  $$ = makeNode($1);
	  $$->attr = $3;
	  if($$->level!=-1 && $$->level != $2)
	    yyerror("two different levels given");
	  else
	    $$->level = $2;
	  if($$->label)
	    yyerror("label given twice");
	  else
	    $$->label = $4;
	}
;

edgeDefList
:
  edgeDef edgeDefList 
	{ $$ = list_cons($1,$2); }
|
  /* epsilon */
    { $$ = 0; }
;

edgeDef
:
  direction IDENT attributeRef
  	{ $$ = newEdgeParameters($1,makeNode($2),$3); }
;

direction
:
  GREATERTHAN
	{ $$ = TO_EDGE; }
|
  LESSTHAN
	{ $$ = FROM_EDGE; }
;

attributeRef
:
  COLON IDENT
	{ $$ = makeAttr($2); }
|
  /* epsilon */
  	{ $$ = default_attr; }
;

levelDef
:
  LEVEL
|
  /* epsilon */
	{ $$ = -1; }
;

labelDef
:
  STRING
	{ $$ = newLabel($1); }
|
  /* epsilon */
  	{ $$ = 0; }
;

attrDef
:
  IDENT LC attrComponentsList RC
  	{
	  makeAttr($1);
	  insertComponents($3,makeAttr($1));
	}
;

attrComponentsList
:
  IDENT attrComponentsList
	{
	  if(!isValidAttrComponent(makeStr($1)))
	    yyerror("nonvalid component of an attribute");
	  $$ = list_cons(makeStr($1),$2);
	}
|
  /* epsilon */
	{ $$ = 0; }
;

%%
int yywrap() { return 1; }
