#include <string.h>

#include "lists.h"
#include "symbols.h"
#include "mem.h"

#include "macros.h"

#define SIZE 92683

symbol hashtab[SIZE];

int hash(char *s)
{
  int h;

  for (h=0; *s; s++)
    h = ((h << 5) | (h>>(8*sizeof(int)-5))) ^ *s;

  return abs(h%SIZE);
}

symbol findSymbol(char *s)
{
  symbol l, *p;
  int h=hash(s);

  for(p = &hashtab[h]; *p; p = &(*p)->next)
    if(strcmp ((*p)->name, s) == 0) return *p;

  l = newCell(sizeof(_symbol));
  l->type = 0;
  l->name = strcpy((char *) malloc (strlen (s) + 1), s);
  l->attr = 0;
  l->def = 0;
  l->next = 0;

  return *p = l;
}

void disposeSymbols()
{
  int i;

  for(i=0; i<SIZE; i++)
    {
      symbol tmp, s;

      for(s=hashtab[i]; s; s=tmp)
        {
	  tmp=s->next;
	  free(s->name);
	  freeCell(s, sizeof(_symbol));
	}
    }
}

int isDefined(symbol s) { return s->type != 0; }
int isNode(symbol s) { return s->type & NODE; }
int isAttr(symbol s) { return s->type & ATTR; }

list unresolved=0;

void partialInsertList(node n, list listin)
{
  list l=listin;

  edge e;

  int dir;
  node dest;
  attribute a;

  edge_parameters ep;

  while(l)
    {
      ep = (edge_parameters)car(l);

      dir = ep->direction;
      dest = ep->destination;
      a = ep->attr;

      freeCell(ep,sizeof(_edge_parameters));

      e = dir == TO_EDGE ? newEdge(n,dest) : newEdge(dest,n);
      e->attr = a;

      l = cdr(l);
    }
  list_freeList(listin);
}

edge_parameters newEdgeParameters(int dir, node dest, attribute a)
{
  edge_parameters tmp = newCell(sizeof(_edge_parameters));

  tmp->direction = dir;
  tmp->destination = dest;
  tmp->attr = a;

  return tmp;
}

node makeNode(symbol s)
{
  if(isNode(s)) return s->def;
  else
    {
      s->type |= NODE;
      return s->def = newNode(s->name);
    }
}

node findNode(char*str)
{
  symbol s = findSymbol(str);
  return makeNode(s);
}

attribute findAttr(char*str)
{
  symbol s = findSymbol(str);
  return makeAttr(s);
}

attribute makeAttr(symbol s)
{
  if(isAttr(s)) return s->attr;
  else
    {
      s->type |= ATTR;
      return s->attr = newAttr(s->name);
    }
}

char* makeStr(symbol s) { return s->name; }

void reverseAllLists() { reverseAllNodes(); reverseAllAttributes(); }
